from linter_adapters.common import BaseCategory, Severity
from linter_adapters.parser import BaseViolationParser


class Flake8BanditCategory(BaseCategory):
    TestCode            = 100
    Misconfiguration    = 200
    BadCalls            = 300
    BadImports          = 400
    Cryptography        = 500
    Injection           = 600
    XSS                 = 700

@BaseViolationParser.register_decorator('S')
class Flake8BanditViolationParser(BaseViolationParser[Flake8BanditCategory]):
    ViolationCategory = Flake8BanditCategory
    provider_name = 'flake8-bandit'
    severity_mapping = \
    {
        Flake8BanditCategory.TestCode:         Severity.Error,
        Flake8BanditCategory.Misconfiguration: Severity.Error,
        Flake8BanditCategory.BadCalls:         Severity.Error,
        Flake8BanditCategory.BadImports:       Severity.Error,
        Flake8BanditCategory.Cryptography:     Severity.Error,
        Flake8BanditCategory.Injection:        Severity.Error,
        Flake8BanditCategory.XSS:              Severity.Error,
    }
    default_severity = Severity.Error


__all__ = \
[
    'Flake8BanditCategory',
    'Flake8BanditViolationParser',
]



"""
Inspired by `flake8_for_pycharm` package.
Original authors:
* Copyright (C) 2017-2018 Ian Stapleton Cordasco <graffatcolmingov@gmail.com> - license holder
* Ramast Magdy - publisher

Author of this module: Peter Zaitcev / USSX-Hares, 2025
"""

import json

from flake8.formatting.base import BaseFormatter
from flake8.violation import Violation

from linter_adapters.common import Severity
from linter_adapters.parser import BaseViolationParser
from .pylint import PylintMessage


_END_OF_OUTPUT = PylintMessage \
(
    type = Severity.Informational.value,
    module = '',
    obj = '',
    line = 1,
    column = 0,
    endColumn = None,
    endLine = None,
    path = '/dev/null',
    symbol = 'eof',
    message = "This is end of file",
    message_id = 'I000',
)

_PYFLAKES_MAP: dict[str, str] | None = None


class PylintJson(BaseFormatter):
    def __init__(self, options):
        self.first_error: bool = True
        super(PylintJson, self).__init__(options)
    
    def write_line(self, line):
        """Override write for convenience."""
        self.write(line, None)
    
    def start(self):
        self.newline = '\n'
        self.write_line('[')
    
    def stop(self):
        if (self.newline != '\n'):
            self.newline = '\n'
            self.write_line(self.format_pylint_message(_END_OF_OUTPUT))
        self.newline = '\n'
        self.write_line(']')
    
    def format_pylint_message(self, message: PylintMessage) -> str:
        indent = ' ' * 4
        formatted = json.dumps(message.asdict(), indent=indent)
        formatted = indent + formatted.replace('\n', '\n' + indent)
        return formatted
    
    def format(self, violation: Violation) -> str:
        """ Format a violation. """
        
        formatted = Flake8ToPylint.convert_violation(violation)
        
        max_line_length = 999
        if (violation.physical_line):
            max_line_length = len(violation.physical_line)
        if (violation.column_number >= max_line_length):
            formatted = formatted._replace(column=max_line_length - 1)
        
        return self.format_pylint_message(formatted)
    
    def handle(self, error: Violation):
        self.newline = ',\n'
        self.write_line(self.format(error))


class Flake8ToPylint:
    @classmethod
    def convert_violation(cls, error: Violation) -> PylintMessage:
        violation_prototype = BaseViolationParser.parse_using_registry(error)
        formatted = PylintMessage \
        (
            type = (violation_prototype.severity or Severity.Convention).value,
            module = BaseViolationParser.parse_module_name(error.filename),
            obj = "",
            line = error.line_number,
            column = error.column_number,
            endColumn = None,
            endLine = None,
            path = error.filename,
            symbol = violation_prototype.string_id or error.code,
            message = f"[{error.code}] {error.text}",
            message_id = error.code,
        )
        
        return formatted


__all__ = \
[
    'PylintJson',
    'Flake8ToPylint',
]



import sys
from argparse import ArgumentParser, Namespace
from typing import *

from flake8.main.cli import main as flake8_main

from linter_adapters.integrations import register_all
from linter_adapters.parser import BaseViolationParser
from linter_adapters.types import ImportableFunction


MainFunction: TypeAlias = Callable[[Sequence[str]], int | None]

class ArgumentsNamespace(Namespace):
    output_format: Literal['text', 'json'] = 'text'
    paths: Sequence[str] | None = None
    help_msg: Sequence[str] | None = None
    flake8_main: ImportableFunction[MainFunction] | None = None

def explain(*error_codes: str) -> int:
    for error_code in error_codes:
        violation_prototype = BaseViolationParser.parse_using_registry(error_code, optimistic=False)
        if (violation_prototype is None):
            print(f"No such message id or symbol '{error_code}'.")
        else:
            print(f":{violation_prototype.string_id} ({violation_prototype.error_code}) from {violation_prototype.provider}")
            if violation_prototype.details:
                print(f"  {violation_prototype.details}")
        print()
    
    return 0

def run(argv: Sequence[str], *, parsed_args: ArgumentsNamespace = None) -> int:
    argv = list(argv)
    parsed_args = parsed_args or ArgumentsNamespace()
    
    match parsed_args.output_format:
        case 'text':
            fmt = 'pylint'
        case 'json':
            fmt = 'pylint-json'
        case _:
            fmt = None
    
    if (fmt is not None):
        argv.insert(0, f'--format={fmt}')
    
    argv.extend(parsed_args.paths or ())
    
    run_flake8 = parsed_args.flake8_main.func if (parsed_args.flake8_main) else flake8_main
    return run_flake8(argv) or 0


def make_argument_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument('--help-msg', nargs='+', default=ArgumentsNamespace.help_msg)
    parser.add_argument('--output-format', '-f', choices=[ 'text', 'json' ], default=ArgumentsNamespace.output_format)
    parser.add_argument('--flake8-main', '-F', type=ImportableFunction[MainFunction], default=ArgumentsNamespace.flake8_main)
    parser.add_argument('paths', type=str, nargs='*', default=ArgumentsNamespace.paths)
    return parser


def main(argv: Sequence[str] = None) -> int:
    if (argv is None):
        argv = sys.argv[1:]
    
    parser = make_argument_parser()
    parsed, unparsed = parser.parse_known_args(argv, namespace=ArgumentsNamespace())
    
    register_all()
    
    if (parsed.help_msg is not None):
        return explain(*parsed.help_msg)
    else:
        return run(unparsed, parsed_args=parsed)

if (__name__ == '__main__'):
    main()


__all__ = \
[
    'ArgumentsNamespace',
    
    'explain',
    'main',
    'make_argument_parser',
    'run',
]



from typing import *


class PylintMessage(NamedTuple):
    type: str
    module: str
    obj: str
    line: int | None
    column: int | None
    endColumn: int | None
    endLine: int | None
    path: str
    symbol: str
    message: str
    message_id: str  # should be `message-id`
    
    def asdict(self) -> dict[str, Any]:
        result = self._asdict()
        result['message-id'] = result.pop('message_id')
        return result


__all__ = \
[
    'PylintMessage',
]



from linter_adapters.common import BaseCategory, Severity
from linter_adapters.parser import BaseViolationParser


class Flake8RSTDocstringsCategory(BaseCategory):
    Info            = 100
    Warning         = 200
    Error           = 300
    Severe          = 400
    ParsingError    = 900

@BaseViolationParser.register_decorator('RST')
class Flake8RSTDocstringsViolationParser(BaseViolationParser[Flake8RSTDocstringsCategory]):
    ViolationCategory = Flake8RSTDocstringsCategory
    provider_name = 'flake8-rst-docstrings'
    severity_mapping = \
    {
        Flake8RSTDocstringsCategory.Info:           Severity.Warning,
        Flake8RSTDocstringsCategory.Warning:        Severity.Warning,
        Flake8RSTDocstringsCategory.Error:          Severity.Error,
        Flake8RSTDocstringsCategory.Severe:         Severity.Error,
        Flake8RSTDocstringsCategory.ParsingError:   Severity.Fatal,
    }
    default_severity = Severity.Warning


__all__ = \
[
    'Flake8RSTDocstringsCategory',
    'Flake8RSTDocstringsViolationParser',
]



from linter_adapters.common import BaseCategory, Severity
from linter_adapters.parser import BaseViolationParser


class Flake8StringFormatCategory(BaseCategory):
    ImplicitParameters  = 100
    MissingValues       = 200
    UnusedValues        = 300

@BaseViolationParser.register_decorator('FMT')
class Flake8StringFormatViolationParser(BaseViolationParser[Flake8StringFormatCategory]):
    ViolationCategory = Flake8StringFormatCategory
    provider_name = 'flake8-string-format'
    severity_mapping = \
    {
        Flake8StringFormatCategory.ImplicitParameters:  Severity.Error,
        Flake8StringFormatCategory.MissingValues:       Severity.Error,
        Flake8StringFormatCategory.UnusedValues:        Severity.Warning,
    }
    default_severity = Severity.Warning


__all__ = \
[
    'Flake8StringFormatCategory',
    'Flake8StringFormatViolationParser',
]



# noinspection PyProtectedMember
def register_all():
    from .checkers.flake8_bandit            import __all__
    from .checkers.flake8_bugbear           import __all__
    from .checkers.flake8_class_c           import __all__
    from .checkers.flake8_docstrings        import __all__
    from .checkers.flake8_oneliners         import __all__
    from .checkers.flake8_rst               import __all__
    from .checkers.flake8_string_format     import __all__
    from .checkers.pycodestyle              import __all__
    from .checkers.pyflakes                 import __all__
    from .checkers.wemake_python_styleguide import __all__



from typing import *

from linter_adapters.common import BaseCategory, Severity, ViolationPrototype
from linter_adapters.parser import BaseViolationParser


try:
    from flake8.plugins.pyflakes import FLAKE8_PYFLAKES_CODES
except ImportError:
    FLAKE8_PYFLAKES_CODES = dict()

class PyflakesCategory(BaseCategory):
    Imports = 400
    FormatStrings = 500
    Misuses = 600
    Statements = 700
    Variables = 800
    CommonMistakes = 900


@BaseViolationParser.register_decorator('F')
class PyflakesViolationParser(BaseViolationParser[PyflakesCategory]):
    provider_name = 'pyflakes'
    ViolationCategory = PyflakesCategory
    severity_mapping = \
    {
        PyflakesCategory.Imports:           Severity.Error,
        PyflakesCategory.FormatStrings:     Severity.Error,
        PyflakesCategory.Misuses:           Severity.Error,
        PyflakesCategory.Statements:        Severity.Error,
        PyflakesCategory.Variables:         Severity.Warning,
        PyflakesCategory.CommonMistakes:    Severity.Error,
    }
    default_severity = Severity.Error
    
    pyflakes_reverse_mapping: Dict[str, str] = { value: key for key, value in FLAKE8_PYFLAKES_CODES.items() }
    
    def parse_violation_extras(self, violation: ViolationPrototype[PyflakesCategory]) -> ViolationPrototype[PyflakesCategory]:
        violation.string_id = self.pyflakes_reverse_mapping.get(violation.error_code)
        return violation


__all__ = \
[
    'PyflakesCategory',
    'PyflakesViolationParser',
]



from typing import Dict

from linter_adapters.common import BaseCategory, Severity, ViolationPrototype
from linter_adapters.parser import BaseViolationParser


class Flake8CCategory(BaseCategory):
    Comprehensions      = 400
    CommasPlacement     = 800
    McCabeComplexity    = 900

@BaseViolationParser.register_decorator('C')
class Flake8ComprehensionOrMcCabeViolationParser(BaseViolationParser[Flake8CCategory]):
    """ Both use error class 'C' """
    
    provider_name = 'comprehensions-or-commas-or-mccabe'
    ViolationCategory = Flake8CCategory
    severity_mapping = \
    {
        Flake8CCategory.Comprehensions:     Severity.Refactor,
        Flake8CCategory.CommasPlacement:    Severity.Convention,
        Flake8CCategory.McCabeComplexity:   Severity.Refactor,
    }
    default_severity = Severity.Warning
    
    provider_names_mapping: Dict[Flake8CCategory, str] = \
    {
        Flake8CCategory.Comprehensions:     'flake8-comprehensions',
        Flake8CCategory.CommasPlacement:    'flake8-commas',
        Flake8CCategory.McCabeComplexity:   'mccabe',
    }
    
    def parse_violation_extras(self, violation: ViolationPrototype[Flake8CCategory]) -> ViolationPrototype[Flake8CCategory]:
        if (violation.category in self.provider_names_mapping):
            violation.provider = self.provider_names_mapping[violation.category]
        
        return violation


__all__ = \
[
    'Flake8CCategory',
    'Flake8ComprehensionOrMcCabeViolationParser',
]



from linter_adapters.common import BaseCategory, Severity, ViolationPrototype
from linter_adapters.parser import BaseViolationParser


class PycodestyleCategory(BaseCategory):
    Indentation = 100
    Whitespace  = 200
    BlankLines  = 300
    Imports     = 400
    LineLength  = 500
    Deprecation = 600
    Statements  = 700
    Comments    = 800   # comes from flake8-eradicate
    SyntaxError = 900


@BaseViolationParser.register_decorator('E', 'W')
class PycodestyleViolationParser(BaseViolationParser[PycodestyleCategory]):
    provider_name = 'pycodestyle'
    ViolationCategory = PycodestyleCategory
    severity_mapping = \
    {
        PycodestyleCategory.Indentation:    Severity.Error,
        PycodestyleCategory.Whitespace:     Severity.Error,
        PycodestyleCategory.BlankLines:     Severity.Convention,
        PycodestyleCategory.Imports:        Severity.Convention,
        PycodestyleCategory.LineLength:     Severity.Convention,
        PycodestyleCategory.Deprecation:    Severity.Warning,
        PycodestyleCategory.Statements:     Severity.Warning,
        PycodestyleCategory.Comments:       Severity.Error,
        PycodestyleCategory.SyntaxError:    Severity.Fatal,
    }
    default_severity = Severity.Warning
    
    def parse_violation_extras(self, violation: ViolationPrototype[PycodestyleCategory]) -> ViolationPrototype[PycodestyleCategory]:
        violation.string_id = violation.category.name if (violation.category is not None) else None
        if (violation.category == PycodestyleCategory.Comments):
            violation.provider = 'flake8-eradicate'
        return violation


__all__ = \
[
    'PycodestyleCategory',
    'PycodestyleViolationParser',
]



from linter_adapters.common import BaseCategory, Severity
from linter_adapters.parser import BaseViolationParser


class Flake8BugbearCategory(BaseCategory):
    ProbablyErrors  = 000
    Optimizations   = 900

@BaseViolationParser.register_decorator('B')
class Flake8BugbearViolationParser(BaseViolationParser[Flake8BugbearCategory]):
    ViolationCategory = Flake8BugbearCategory
    provider_name = 'flake8-bugbear'
    severity_mapping = \
    {
        Flake8BugbearCategory.ProbablyErrors:   Severity.Error,
        Flake8BugbearCategory.Optimizations:    Severity.Refactor,
    }
    default_severity = Severity.Error


__all__ = \
[
    'Flake8BugbearCategory',
    'Flake8BugbearViolationParser',
]



from linter_adapters.common import Severity
from linter_adapters.parser import SingleCategoryViolationParser


@SingleCategoryViolationParser.register_decorator('T')
class Flake8DebuggerViolationParser(SingleCategoryViolationParser):
    provider_name = 'flake8-debugger'
    category_name = 'DebuggerUsage'
    category_number = 100
    default_severity = Severity.Error


@SingleCategoryViolationParser.register_decorator('I')
class Flake8ImportSortViolationParser(SingleCategoryViolationParser):
    provider_name = 'flake8-isort'
    category_name = 'ImportsPositioning'
    category_number = 000
    default_severity = Severity.Convention


@SingleCategoryViolationParser.register_decorator('N')
class Flake8BrokenLineViolationParser(SingleCategoryViolationParser):
    provider_name = 'flake8-broken-line'
    category_name = 'Linebreaks'
    category_number = 400
    default_severity = Severity.Convention


@SingleCategoryViolationParser.register_decorator('DTZ')
class Flake8DateTimeZoneViolationParser(SingleCategoryViolationParser):
    provider_name = 'flake8-datetimez'
    category_name = 'Timezone'
    category_number = 000
    default_severity = Severity.Error


@SingleCategoryViolationParser.register_decorator('Q')
class Flake8QuotesViolationParser(SingleCategoryViolationParser):
    provider_name = 'flake8-quotes'
    category_name = 'Quotes'
    category_number = 000
    default_severity = Severity.Convention


__all__ = \
[
    'Flake8BrokenLineViolationParser',
    'Flake8DateTimeZoneViolationParser',
    'Flake8DebuggerViolationParser',
    'Flake8ImportSortViolationParser',
    'Flake8QuotesViolationParser',
]



from collections import namedtuple


__title__ = 'linter-adapters'
__author__ = 'Peter Zaitcev'
__license__ = 'BSD 2-clause license'
__copyright__ = 'Copyright 2025 Peter Zaitcev'
__version__ = '0.1.0'

VersionInfo = namedtuple('VersionInfo', 'major minor micro releaselevel serial')
version_info = VersionInfo(*__version__.split('.'), releaselevel='alpha', serial=0)

__all__ = \
[
    'version_info',
    '__title__',
    '__author__',
    '__license__',
    '__copyright__',
    '__version__',
]



from linter_adapters.common import BaseCategory, Severity
from linter_adapters.parser import BaseViolationParser


class Flake8DocstringsCategory(BaseCategory):
    MissingDocstrings       = 100
    WhitespaceIssues        = 200
    QuotesIssues            = 300
    DocstringContentIssues  = 400
    DocstringParsingError   = 900

@BaseViolationParser.register_decorator('D')
class Flake8DocstringsViolationParser(BaseViolationParser[Flake8DocstringsCategory]):
    ViolationCategory = Flake8DocstringsCategory
    provider_name = 'flake8-docstrings'
    severity_mapping = \
    {
        Flake8DocstringsCategory.MissingDocstrings:         Severity.Warning,
        Flake8DocstringsCategory.WhitespaceIssues:          Severity.Warning,
        Flake8DocstringsCategory.QuotesIssues:              Severity.Warning,
        Flake8DocstringsCategory.DocstringContentIssues:    Severity.Error,
        Flake8DocstringsCategory.DocstringParsingError:     Severity.Fatal,
    }
    default_severity = Severity.Warning


__all__ = \
[
    'Flake8DocstringsCategory',
    'Flake8DocstringsViolationParser',
]



import os
from dataclasses import dataclass, field
from functools import cached_property
from importlib.metadata import EntryPoint
from typing import *


AnyPath = str | os.PathLike[str]

@dataclass
class ImportableObject[T]:
    _entry_point: EntryPoint = field(init=False, repr=False)
    
    def __init__(self, name: str):
        self._entry_point = EntryPoint(name=None, group=None, value=name)
    
    @cached_property
    def obj(self) -> T:
        return self._entry_point.load()


class ImportableFunction[C: Callable](ImportableObject[C]):
    @property
    def func(self) -> C:
        return self.obj


__all__ = \
[
    'AnyPath',
    'ImportableFunction',
    'ImportableObject',
]



from __future__ import annotations

import typing
from typing import *

from linter_adapters.common import BaseCategory, Severity, ViolationPrototype
from linter_adapters.parser import BaseViolationParser


if (typing.TYPE_CHECKING):
    from wemake_python_styleguide.cli.commands.explain.violation_loader import ViolationInfo

try:
    from wemake_python_styleguide.cli.commands.explain.violation_loader import get_violation
except ImportError:
    get_violation = lambda code: None


class WemakePythonStyleguideCategory(BaseCategory):
    System = 000
    """
    These checks ensures that our internal checks passes.
    
    For example, we can report violations from this group
    when some exception occur during the linting process
    or some dependencies are missing.
    """
    
    Naming = 100
    """
    Naming is hard! It is, in fact, one of the two hardest problems.

    These checks are required to make your application
    easier to read and understand by multiple people over the long period of time.
    """
    
    Complexity = 200
    """
    These checks find flaws in your application design.
    
    We try to stick to “the magical 7 ± 2 number” when counting things.
    https://en.wikipedia.org/wiki/The_Magical_Number_Seven,_Plus_or_Minus_Two
    
    That’s how many objects we can keep in our memory at a time.
    We try hard not to exceed the memory capacity limit.
    
    You can also find interesting reading about “Cognitive complexity”:
    https://www.sonarsource.com/docs/CognitiveComplexity.pdf
    """
    
    Consistency = 300
    """
    These checks limit Python’s inconsistencies.
    
    We can do the same things differently in Python.
    For example, there are three ways to format a string.
    There are several ways to write the same number.
    
    We like our code to be consistent.
    It is easier to work with your code base if you follow these rules.
    
    So, we choose a single way to do things.
    It does not mean that we choose the best way to do it.
    But, we value consistency more than being 100% right
    and we are ready to suffer all trade-offs that might come.
    
    Once again, these rules are highly subjective, but we love them.
    """
    
    BestPractices = 400
    """
    These checks ensure that you follow the best practices.
    
    The source for these best practices is countless hours
    we have spent debugging software or reviewing it.
    
    How do we find inspiration for new rules?
    We find some ugly code during code reviews and audits,
    then we forbid the use of code like it forever.
    """
    
    Refactoring = 500
    """
    These checks ensure that you don’t have patterns that can be refactored.
    
    There are so many ways of doing the same thing in Python.
    Here we collect know patterns that can be rewritten
    into much easier or just more pythonic version.
    """
    
    OOP = 600
    """
    These checks ensures that you use Python’s version of OOP correctly.
    
    There are different gotchas in Python to write beautiful classes and using objects correctly.
    That’s the place we collect these kind of rules.
    """


@BaseViolationParser.register_decorator('WPS')
class WemakePythonStyleguideViolationParser(BaseViolationParser[WemakePythonStyleguideCategory]):
    provider_name = 'wemake-python-styleguide'
    ViolationCategory = WemakePythonStyleguideCategory
    severity_mapping = \
    {
        WemakePythonStyleguideCategory.System:          Severity.Fatal,
        WemakePythonStyleguideCategory.Naming:          Severity.Convention,
        WemakePythonStyleguideCategory.Complexity:      Severity.Refactor,
        WemakePythonStyleguideCategory.Consistency:     Severity.Warning,
        WemakePythonStyleguideCategory.BestPractices:   Severity.Convention,
        WemakePythonStyleguideCategory.Refactoring:     Severity.Refactor,
        WemakePythonStyleguideCategory.OOP:             Severity.Warning,
    }
    default_severity = Severity.Convention
    
    def get_wemake_python_violation_info(self, violation: ViolationPrototype) -> Optional[ViolationInfo]:
        return get_violation(violation.error_code_number)
    
    def parse_violation_extras(self, violation: ViolationPrototype[WemakePythonStyleguideCategory]) -> ViolationPrototype[WemakePythonStyleguideCategory]:
        violation_info = self.get_wemake_python_violation_info(violation)
        violation.string_id = violation_info.identifier if violation_info else None
        violation.details = violation_info
        
        return violation


__all__ = \
[
    'WemakePythonStyleguideCategory',
    'WemakePythonStyleguideViolationParser',
]



from extended_setup import ExtendedSetupManager

ExtendedSetupManager('linter_adapters').setup \
(
    short_description = "An adapter which provides categorization & descriptions on popular linters & checkers",
    category = 'tools',
    min_python_version = '3.10',
    classifiers =
    [
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Framework :: Flake8',        
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
    ],
    entry_points =
    {
        'flake8.report':    [ 'pylint-json = linter_adapters.integrations.formatters.flake8_as_pylint_json:PylintJson' ],
        'console_scripts':  [ 'flake8_as_pylint = linter_adapters.integrations.formatters.flake8_for_pycharm:main' ],
    }
)



from __future__ import annotations

import re
import typing
from abc import ABC
from pathlib import Path
from typing import *

from .common import BaseCategory, Severity, ViolationPrototype
from .types import AnyPath


if (typing.TYPE_CHECKING):
    from flake8.violation import Violation as Flake8Violation

AnyViolation: TypeAlias = Union[str, 'Flake8Violation', ViolationPrototype[Any]]

class BaseViolationParser[C: BaseCategory]:
    REGISTRY: ClassVar[Dict[str, BaseViolationParser]] = dict()
    
    provider_name: str
    ViolationCategory: Type[C]
    severity_mapping: Dict[C, Severity]
    
    default_severity: Severity = Severity.Warning
    violation_code_number_base: int = 100
    
    @classmethod
    def parse_module_name(cls, filename: AnyPath):
        return Path(filename) \
            .absolute() \
            .relative_to(Path('.').absolute()) \
            .with_suffix('') \
            .as_posix() \
            .replace('/', '.') \
    
    def parse_violation_base(self, error_code: str) -> ViolationPrototype[None] | None:
        parsed = re.fullmatch(r'(?P<provider>\w+?)(?P<number>\d+)', error_code)
        if not parsed:
            return None
        
        result: ViolationPrototype[None] = ViolationPrototype(error_code, parsed.group('provider'), int(parsed.group('number')))
        return result
    
    def parse_violation_category(self, violation: ViolationPrototype[Any]) -> ViolationPrototype[C]:
        violation.provider = self.provider_name
        
        try:
            violation.category = self.ViolationCategory(violation.error_code_number // self.violation_code_number_base * self.violation_code_number_base)
        except (ValueError, LookupError):
            violation.category = None
        
        violation.severity = self.severity_mapping.get(violation.category, self.default_severity)
        
        return violation
    
    def parse_violation_extras(self, violation: ViolationPrototype[C]) -> ViolationPrototype[C]:
        return violation
    
    def parse_violation_code(self, violation: AnyViolation) -> str:
        match violation:
            case str(code):
                error_code = code
            case ViolationPrototype(error_code=code):
                error_code = code
            case flake8_violation if flake8_violation.__class__.__name__ == 'Violation':  # type: Flake8Violation
                error_code = flake8_violation.code
            case _:
                error_code = str(violation)
        
        return error_code
    
    def parse_violation(self, violation: AnyViolation) -> ViolationPrototype[C] | None:
        error_code = self.parse_violation_code(violation)
        violation = violation and self.parse_violation_base(error_code)
        violation = violation and self.parse_violation_category(violation)
        violation = violation and self.parse_violation_extras(violation)
        return violation
    
    @classmethod
    def parse_using_registry(cls, violation: AnyViolation, *, optimistic: bool = True) -> ViolationPrototype[Any] | None:
        parser = SimpleViolationParser.INSTANCE
        error_code = parser.parse_violation_code(violation)
        violation = violation and parser.parse_violation_base(error_code)
        parser = BaseViolationParser.REGISTRY.get(violation.error_code_class, parser if (optimistic) else None)
        return parser and parser.parse_violation(violation)
    
    @classmethod
    def register(cls, *error_classes: str):
        instance = cls()
        instance.register_violation_class(*error_classes)
    
    @classmethod
    def register_decorator[T: BaseViolationParser](cls, *error_classes: str) -> Callable[[Type[T]], Type[T]]:
        def decorator(parser_cls: Type[T]) -> Type[T]:
            parser_cls.register(*error_classes)
            return parser_cls
        return decorator
    
    def register_violation_class(self, *error_classes: str) -> None:
        for error_class in error_classes:
            assert error_class not in self.REGISTRY, f"Cannot re-register error class '{error_class}'"
            BaseViolationParser.REGISTRY[error_class] = self


class SingleCategoryViolationParser[C: BaseCategory](BaseViolationParser, ABC):
    category_name: str
    category_number: int
    
    @classmethod
    def make_enum(cls) -> Type[C]:
        enum_name = cls.__name__.replace('ViolationParser', '') + 'Category'
        enum_cls = BaseCategory(enum_name, { cls.category_name: cls.category_number })
        return cast(Type[C], enum_cls)
    
    def __init_subclass__(cls, **kwargs):
        if (not hasattr(cls, 'ViolationCategory')):
            cls.ViolationCategory = cls.make_enum()
        if (not hasattr(cls, 'severity_mapping')):
            cls.severity_mapping = { cls.ViolationCategory[cls.category_name]: cls.default_severity }
        
        super().__init_subclass__()


class SimpleViolationParser(SingleCategoryViolationParser):
    provider_name = 'linter-adapters'
    category_name = 'Violation'
    category_number = 000

SimpleViolationParser.INSTANCE = SimpleViolationParser()


__all__ = \
[
    'BaseViolationParser',
    'SingleCategoryViolationParser',
    'SimpleViolationParser',
]



from dataclasses import dataclass
from enum import IntEnum, StrEnum
from typing import *


class Severity(StrEnum):
    """ Basically, severity from Pylint. It just makes sense. """
    
    Refactor = 'refactor'
    """ For a "good practice" metric violation """
    
    Convention = 'convention'
    """ For coding standard violation """
    
    Warning = 'warning'
    """ For stylistic problems, or minor programming issues """
    
    Error = 'error'
    """ For important programming issues (i.e. most probably bug) """
    
    Fatal = 'fatal'
    """ For errors which prevented further processing """
    
    Informational = 'info'
    """ Messages that linter emits (do not contribute to your analysis score) """


class BaseCategory(IntEnum):
    pass


@dataclass
class ViolationPrototype[C: BaseCategory]:
    error_code: str
    error_code_class: str
    error_code_number: int
    
    severity: Severity = None # will be present after pre-parsing
    provider: str | None = None
    string_id: str | None = None
    category: C | None = None
    details: Any = None


__all__ = \
[
    'Severity',
    'BaseCategory',
    'ViolationPrototype',
]
